//
//  ContentView.swift
//  ScratchpadApp
//
//  Created by Jasdev Singh on 8/27/20.
//

import Combine
import ComposableArchitecture
import Speech
import SwiftUI

func subRipTimingLine(start: TimeInterval, duration: TimeInterval) -> String {
	"\(timeIntervalToSubRipFormattedTimestamp(start)) --> \(timeIntervalToSubRipFormattedTimestamp(start + duration))"
}

func timeIntervalToSubRipFormattedTimestamp(_ time: TimeInterval) -> String {
	let millisecondsDigits = 3

	let milliseconds = "\(Int((time.truncatingRemainder(dividingBy: 1) * 1_000).rounded()))"
		.prefix(millisecondsDigits)

	let zeroPadding = Array(
		repeating: "0",
		count: millisecondsDigits - milliseconds.count
	)
	.joined()

	return subRipTimeIntervalFormatter.string(from: time)! +
		",\(zeroPadding + milliseconds)"
}

private let subRipTimeIntervalFormatter: DateComponentsFormatter = {
	let formatter = DateComponentsFormatter()
	formatter.zeroFormattingBehavior = .pad
	formatter.allowedUnits = [.hour, .minute, .second]
	return formatter
}()

func transcriptionSegmentsToSubRipText(
	_ segments: [SFTranscriptionSegment]
) -> String {
	segments
		.enumerated()
		.map { sequnceNumber, segment in
			[
				"\(sequnceNumber + 1)", // Shifting the offset up by one, since
				// `EnumeratedSequence.Iterator.Element` starts at zero.
				subRipTimingLine(start: segment.timestamp, duration: segment.duration),
				segment.substring
			]
			.joined(separator: "\n")
		}
		.joined(separator: "\n\n")
}

let transcriptionEffect: (URL) -> Effect<String, NSError> = { audioURL in
	.future { callback in
		guard
			let speechRecognizer = SFSpeechRecognizer(),
			speechRecognizer.isAvailable
		else {
			callback(
				.failure(
					NSError(
						domain: "com.jasdev.CaptionsGeneration",
						code: 1,
						userInfo: nil
					)
				)
			)
			return
		}

		let request = SFSpeechURLRecognitionRequest(url: audioURL)
		request.shouldReportPartialResults = false

		speechRecognizer.recognitionTask(with: request) { [weak request] result, error in
			if let error = error {
				callback(.failure(error as NSError))
			} else {
				callback(
					.success(
						transcriptionSegmentsToSubRipText(
							result!
								.bestTranscription
								.segments
						)
					)
				)
			}
		}
	}
}

struct State: Equatable {
	var transcription = ""
	var isFetchingTranscription = false
}

enum Action {
	case onAppear
	case setTranscriptionString(String)
	case transcriptionResult(Result<String, NSError>)
	case fetchTranscription
}

let transcriptionReducer = Reducer<State, Action, Void> { state, action, _ in
	switch action {
	case .fetchTranscription:
		state.transcription.removeAll()
		state.isFetchingTranscription = true

		return transcriptionEffect(Bundle.main.url(forResource: "sample", withExtension: "m4a")!)
			.subscribe(on: DispatchQueue.global(qos: .userInitiated))
			.catchToEffect()
			.receive(on: DispatchQueue.main)
			.map(Action.transcriptionResult)
			.eraseToEffect()
	case .onAppear:
		return Effect.fireAndForget {
			SFSpeechRecognizer.requestAuthorization { _ in }
		}
	case let .setTranscriptionString(transcription):
		// No-op to disallow editing for demo purposes.
		return .none
	case let .transcriptionResult(result):
		state.isFetchingTranscription = false

		switch result {
		case let .success(transcription):
			state.transcription = transcription
			return .none
		case let .failure(error):
			print(error)
			return .none
		}
	}
}

struct ContentView: View {
	let store: Store<State, Action>

	var body: some View {
		WithViewStore(store) { viewStore in
			VStack {
				TextEditor(
					text: viewStore.binding(
						get: \.transcription,
						send: Action.setTranscriptionString
					)
				)
				.border(Color.primary)
				.clipShape(Rectangle())
				.frame(height: 400)
				.overlay(
					ProgressView()
						.opacity(viewStore.isFetchingTranscription ? 1 : 0)
				)

				Spacer()

				Button("Fetch transcription") { viewStore.send(.fetchTranscription) }
					.opacity(viewStore.isFetchingTranscription ? 0.2 : 1)
					.disabled(viewStore.isFetchingTranscription)
					.padding()
			}
			.foregroundColor(.primary)
			.onAppear { viewStore.send(.onAppear) }
		}
		.background(
			Color.white // Non-`.clear` color here to allow for hit testing.
				.opacity(0.000000000000001)
				.onTapGesture { hideKeyboard() }
		)
	}
}

private extension View {
	func hideKeyboard() {
		UIApplication.shared
			.sendAction(#selector(UIResponder.resignFirstResponder), to: nil, from: nil, for: nil)
	}
}
